var DEBUG = false;
/**** Class Settings (Angular Service) ****/
function Settings() {
  this.enforceLifepathReqts = true;
  this.enforcePointLimits = true;
  this.displayEmotionalMath = false;
}
/**** End Class Settings ****/

/**** Class AppropriateWeaponsService (Angular Service) ****/
function AppropriateWeaponsService($modal, $http) {
  // This class will store a hash which maps lifepath names to a list of
  // weapons that are appropriate for that lifepath.
  this.appropriateWeapons = {};

  var myself = this;

  this.loadFromServer = function(){
    if(serverSettings.storageType != 'server'){
      return;
    }

    $http.get("/get_approp_weapons/user1", {'timeout': 3000} ).
      success(function(data,status,headers,config){
        if (data){
          myself.appropriateWeapons = data
        }
        console.log("Loaded saved appropriate weapons");
      }).
      error(function(data,status,headers,config){
        console.log("Error: Loading appropriate weapons from server failed: HTTP code " + status + ": " + data);
      });
  }

  this.saveToServer = function(){
    if(serverSettings.storageType != 'server'){
      return;
    }

    var json = angular.toJson(myself.appropriateWeapons, true);
    $http.post("/update_approp_weapons/user1", json).
      success(function(data,status,headers,config){
        console.log("Saved appropriate weapons");
      }).
      error(function(data,status,headers,config){
        console.log("Error: Saving appropriate weapons to server failed: HTTP code " + status + ": " + data);
      });
  }

  this.loadFromServer();

  // Replace the 'Appropriate Weapons' entry on the lifepath with the given list.
  this.replaceAppropriateWeapons = function (displayLp, replacementList){
    var newList = [];
    if ( ! replacementList ){
      // There is no replacement for Appropriate Weapons, so leave it alone.
      return;
    }

    var replacementHash = listToHash(replacementList);

    for(var i = 0; i < displayLp.skills.length; i++){
      var skillName = displayLp.skills[i];
      if(skillName != 'Appropriate Weapons' && ! replacementHash[skillName])
      {
        newList.push(displayLp.skills[i]);
      }
    }
    displayLp.skills = newList.concat(replacementList);
  }

  this.replaceAppropriateWeaponsUsingSaved = function (displayLp){
    this.replaceAppropriateWeapons(displayLp, this.appropriateWeapons[displayLp.name]);
  }

  this.hasAppropriateWeapons = function (displayLp){
    var has = false;
    for(var i = 0; i < displayLp.skills.length; i++){
      if(displayLp.skills[i] == 'Appropriate Weapons')
      {
        has = true;
        break;
      }
    }
    return has;
  }

  var myself = this;

  this.selectAppropriateWeapons = function (displayLp, onSelect){
    if( this.hasAppropriateWeapons(displayLp) ){
      this.selectAppropriateWeaponsByLifepathName(displayLp.name, function(selected){
        myself.replaceAppropriateWeapons(displayLp, selected);

        if ( onSelect ){
          onSelect();
        }
      })
    }
  }

  this.selectAppropriateWeaponsByLifepathName = function (lifepathName, onSelect){
    var modalInstance = $modal.open({
      templateUrl: '/choose_appropriate_weapons_partial',
      controller: AppropriateWeaponsModalCtrl,
      resolve: {
        lifepathName: function () {
          return lifepathName;
        },
        chosen: function () {
          if ( myself.appropriateWeapons[lifepathName] )
            return myself.appropriateWeapons[lifepathName];
          else
            return [];
        }
      }
    });

    modalInstance.result.then(function (selected) {
      console.log("Modal: User selected:");
      console.log(selected);
      myself.appropriateWeapons[lifepathName] = selected;
      myself.saveToServer();
      if ( onSelect ){
        onSelect(selected);
      }
    }, function () {
      console.log("Modal: User cancelled");
    });
  }
}
/**** End Class AppropriateWeaponsService ****/

// Replace the 'Weapon of Choice' entry on the lifepath with the given weapon.
function replaceWeaponOfChoice(displayLp, weapon){
  var newList = [];

  if ( ! weapon ){
    // There is no replacement, so leave it alone.
    return;
  }

  for(var i = 0; i < displayLp.skills.length; i++){
    var skillName = displayLp.skills[i];
    if(skillName != 'Weapon Of Choice' && skillName != weapon)
    {
      newList.push(displayLp.skills[i]);
    }
  }
  newList.push(weapon);
  displayLp.skills = newList;
}

/**** Class WeaponOfChoiceService (Angular Service) ****/
function WeaponOfChoiceService($modal, $http) {

  var myself = this;

  this.hasWeaponOfChoice = function (displayLp){
    var has = false;
    for(var i = 0; i < displayLp.skills.length; i++){
      if(displayLp.skills[i] == 'Weapon Of Choice')
      {
        has = true;
        break;
      }
    }
    return has;
  }

  this.selectWeaponOfChoice = function (displayLp, onSelect){
    if( this.hasWeaponOfChoice(displayLp) ){
      this.selectWeaponOfChoiceByModal(displayLp.name, function(selected){
        displayLp.weaponOfChoice = selected;
        replaceWeaponOfChoice(displayLp, selected);

        if ( onSelect ){
          onSelect();
        }
      })
    }
  }

  this.selectWeaponOfChoiceByModal = function (lifepathName, onSelect){
    var modalInstance = $modal.open({
      templateUrl: '/choose_weapon_of_choice_partial',
      controller: WeaponOfChoiceModalCtrl,
      resolve: {
        lifepathName: function () {
          return lifepathName;
        }
      }
    });

    modalInstance.result.then(function (selected) {
      console.log("Modal: User selected:");
      console.log(selected);
      if ( onSelect ){
        onSelect(selected);
      }
    }, function () {
      console.log("Modal: User cancelled");
    });
  }


}

/**** End Class WeaponOfChoiceService ****/

/**** Class CharacterStorageService ****/
function CharacterStorageService($http) {
  this.currentCharacter = null;

  this.characterIdAndNames = [];

  myself = this;

  /* Load character names from server */
  this.loadCharacterNames = function(){
    fetch("/list_chars/user1")
      .then((response) => response.json())
      .then((data) => {
        myself.characterIdAndNames = data;
        console.log("Loaded saved character names");
      })
      .catch((error) => {
        console.log("Error: Loading saved character names from server failed: "+error);
      });
  }

  this.loadCharacterNames();
}

/**** End Class CharacterStorageService ****/

/**** Class BurningDataService (Angular Service) ****/
// This service is used to load the lifepaths, skills, traits, etc. from the server.
function BurningDataService($http) {

  // Used to reference the object from within functions and callbacks
  var myself = this;

   /* JSON Data structure representing lifepaths. The structure is:
      stock:
        setting_name:
          lifepath_name:
            time: N
          ...
   */
  this.lifepaths = {};

  /* JSON data structure representing all available skills. The structure is:
      skill_name:
        roots: [root1, root2, ...]
      skill_name:
        roots: [root1, root2, ...]

  */
  this.skills = {};

  /* JSON data structure representing all available traits */
  this.traits = {};

  /* JSON data structure representing all available resources (gear/property) */
  this.resources = {};

  // A hash of StartingStatPoints objects keyed by stock.
  this.startingStatPts = {};

  /* Loading of stocks, skills, and traits begins on initializing the service

  /* Load stocks from server */
  this.whenStocksLoaded = fetch("/stocks")
    .then((response) => response.json())
    .then((data) => {
      if(DEBUG) {
        console.log("Loaded stock data:");
        console.log(data);
      }
      myself.stocks = data;
      for (var stock of Object.keys(data)) {
        myself.startingStatPts[stock] = new StartingStatPoints(myself.stocks[stock].starting_stats);
      }
    })
    .catch((error) => {
      console.log("Error: Getting stocks from server failed: "+error);
    });

  /* Load skills from server */
  this.whenSkillsLoaded = fetch("/skills")
    .then((response) => response.json())
    .then((data) => {
      myself.skills = data;
      if(DEBUG) {
        console.log("Loaded skill data:");
        console.log(data);
      }
    })
    .catch((error) => {
      console.log("Error: Getting skills from server failed: "+error);
    });

  /* Load traits from server */
  this.whenTraitsLoaded = fetch("/traits")
    .then((response) => response.json())
    .then((data) => {
      myself.traits = data;
      if(DEBUG) {
        console.log("Loaded trait data:");
        console.log(data);
      }
    })
    .catch((error) => {
      console.log("Error: Getting traits from server failed: "+error);
    });

  /* Lifepaths and resources defer until their stock is selected */

  this.whenLifePathsLoadedForStock = {};
  /* Load lifepaths from server */
  this.loadLifepathsForStock = function(stock){
    return myself.whenLifePathsLoadedForStock[stock] = fetch("/lifepaths/" + stock)
      .then((response) => response.json())
      .then((data) => {
        myself.lifepaths[stock] = data;
        if(DEBUG) {
          console.log("Loaded "+stock+" lifepaths:");
          console.log(data);
        }
      })
      .catch((error) => {
        console.log("Error: Getting "+stock+" lifepaths from server failed: "+error);
      });
  };

  /* Load resources from server */
  this.whenResourcesLoadedForStock = {};
  this.loadResourcesForStock = function(stock){
    return myself.whenResourcesLoadedForStock[stock] = fetch("/resources/" + stock)
      .then((response) => response.json())
      .then((data) => {
        myself.resources[stock] = data;
        if(DEBUG) {
          console.log("Loaded "+stock+" resources:");
          console.log(data);
        }
      })
      .catch((error) => {
        console.log("Error: Getting "+stock+" resources from server failed: "+error);
      });
  };
}
/**** End BurningDataService ****/

